from __future__ import annotations

import functools
from typing import overload

import numpy as np
import pandas as pd
from django.db import models
from django.db.models import CASCADE, PROTECT
from lamin_utils import logger
from lamindb.base.fields import (
    BigIntegerField,
    BooleanField,
    CharField,
    ForeignKey,
    TextField,
)
from lamindb.errors import DoesNotExist, InvalidArgument
from lamindb.models import (
    Artifact,
    BaseSQLRecord,
    CanCurate,
    Feature,
    HasParents,
    IsLink,
    Record,
    Schema,
    SQLRecord,
    TracksRun,
    TracksUpdates,
)

import bionty.base as bt_base
from bionty.base.dev._doc_util import _doc_params

from . import ids
from ._biorecord import encode_uid, lookup2kwargs
from ._shared_docstrings import doc_from_source
from .base import PublicOntology
from .base._public_ontology import InvalidParamError


class StaticReference(PublicOntology):
    def __init__(self, source_record: Source) -> None:
        self._source_record = source_record
        super().__init__(
            source=source_record.name,
            version=source_record.version,
            organism=source_record.organism,
        )

    def _load_df(self) -> pd.DataFrame:
        if self._source_record.dataframe_artifact_id:
            self._filter_prefix = False
            return self._source_record.dataframe_artifact.load(is_run_input=False)
        else:
            return pd.DataFrame()


def pass_to_super(cls_method):
    """Decorator to pass all non-None parameters to the superclass method."""

    @functools.wraps(cls_method)
    def wrapper(cls, *args, **kwargs):
        # Get all non-None kwargs
        filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
        # Call superclass method with filtered kwargs
        return getattr(super(cls, cls), cls_method.__name__)(*args, **filtered_kwargs)

    return wrapper


class Source(SQLRecord, TracksRun, TracksUpdates):
    """Versions of ontology sources.

    .. warning::

        Do not modify the records unless you know what you are doing!
    """

    class Meta(SQLRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("entity", "name", "organism", "version"),)

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, default=ids.source)
    """A universal id (base62-encoded hash of defining fields)."""
    entity: str = CharField(max_length=256, db_index=True)
    """Entity class name with schema, e.g. bionty.CellType."""
    organism: str = CharField(max_length=64, db_index=True)
    """Organism name, use 'all' if unknown or none applied."""
    name: str = CharField(max_length=64, db_index=True)
    """Source name, short form, CURIE prefix for ontologies."""
    version: str = CharField(max_length=64, db_index=True)
    """Version of the source."""
    in_db: bool = BooleanField(default=False, db_index=True)
    """Whether this ontology has been added to the database."""
    currently_used: bool = BooleanField(default=False, db_index=True)
    """Whether this record is currently used."""
    description: str | None = TextField(blank=True, db_index=True)
    """Source full name, long form."""
    url: str | None = TextField(null=True, default=None)
    """URL of the source file."""
    md5: str | None = TextField(null=True, default=None)
    """Hash md5 of the source file."""
    source_website: str | None = TextField(null=True, default=None)
    """Website of the source."""
    dataframe_artifact: Artifact = ForeignKey(
        Artifact, PROTECT, null=True, default=None, related_name="_source_dataframe_of"
    )
    """Dataframe artifact that corresponds to this source."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, related_name="_source_artifact_of"
    )
    """Additional files that correspond to this source."""

    @overload
    def __init__(
        self,
        entity: str,
        organism: str,
        name: str,
        version: str,
        currently_used: bool,
        description: str | None,
        url: str | None,
        md5: str | None,
        source_website: str | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        kwargs = encode_uid(registry=Source, kwargs=kwargs)
        super().__init__(*args, **kwargs)

    def save(self, *args, **kwargs) -> Source:
        """Save the source record."""
        update = self.currently_used and self.pk
        super().save(*args, **kwargs)
        # when update currently_used, set all other records of the same source as not currently used
        if update:
            Source.filter(
                entity=self.entity, organism=self.organism, name=self.name
            ).exclude(id=self.id).update(currently_used=False)
        return self


class BioRecord(SQLRecord, HasParents, CanCurate):
    """Base SQLRecord of bionty.

    BioRecord inherits all methods from :class:`~lamindb.models.SQLRecord` and provides additional methods
    including :meth:`~bionty.core.BioRecord.public` and :meth:`~bionty.core.BioRecord.from_source`.

    Notes:
        For more info, see tutorials:

        - :doc:`docs:bionty`
        - :doc:`docs:manage-ontologies`
    """

    class Meta:
        abstract = True

    source: Source = ForeignKey(Source, PROTECT, null=True, related_name="+")
    """:class:`~bionty.Source` this record associates with."""

    def __init__(self, *args, **kwargs):
        # DB-facing constructor
        if len(args) == len(self._meta.concrete_fields):
            super().__init__(*args, **kwargs)
            return None

        # passing lookup result from bionty, which is a Tuple or List
        if (
            args
            and len(args) == 1
            and isinstance(args[0], tuple | list)
            and len(args[0]) > 0
        ):
            if isinstance(args[0], list) and len(args[0]) > 1:
                logger.warning(
                    "multiple lookup/search results were passed. Only returning record from the first entry."
                )
            result = lookup2kwargs(self, *args, **kwargs)  # type:ignore
            # exclude "parents" from query arguments
            query_kwargs = {k: v for k, v in result.items() if k != "parents"}
            existing_record = self.__class__.filter(**query_kwargs).one_or_none()
            if existing_record is not None:
                from lamindb.models.sqlrecord import init_self_from_db

                init_self_from_db(self, existing_record)
                return None
            else:
                kwargs = result  # result already has encoded id
                args = ()

        # raise error if no organism is passed
        if hasattr(self.__class__, "organism_id"):
            if kwargs.get("organism") is None and kwargs.get("organism_id") is None:
                from .core._settings import settings

                if settings.organism is not None:
                    kwargs["organism"] = settings.organism
                else:
                    raise RuntimeError("please pass a organism!")
            elif kwargs.get("organism") is not None:
                if not isinstance(kwargs.get("organism"), Organism):
                    raise TypeError("organism must be a `bionty.Organism` record.")

        kwargs = encode_uid(registry=self.__class__, kwargs=kwargs)

        # now continue with the user-facing constructor
        # set the direct parents as a private attribute
        # this is a list of strings that store the ontology id
        if "parents" in kwargs:
            parents = kwargs.pop("parents")
            # this checks if we receive a np.ndarray from pandas
            if isinstance(parents, list | np.ndarray) and len(parents) > 0:
                if not isinstance(parents[0], str):
                    raise ValueError(
                        "Invalid parents kwarg passed. Provide a list of ontology ids."
                    )
                self._parents = parents

        super().__init__(*args, **kwargs)

    @classmethod
    def import_source(
        cls,
        source: Source | None = None,
        update_records: bool = False,
        *,
        organism: str | SQLRecord | None = None,
        ignore_conflicts: bool = True,
    ):
        """Bulk save records from a Bionty ontology.

        Use this method to initialize your registry with public ontology.

        Args:
            source: Source record to import records from.
            update_records: If True, update existing records with the new source.

                - If a record has the same metadata in the new source, link the record to the new source.
                - If a record has no artifacts associated, update it's metadata and link to the new source.
                - If a record associated artifacts, but different name in the new source, create a new record with the new source.

            organism: Organism name or record.
                Required for entities with a required organism foreign key when no source is passed.
            ignore_conflicts: Whether to ignore conflicts during bulk record creation.

        Examples::

            import bionty as bt

            # import all records from a default source
            default_sources = bt.Source.filter(entity="bionty.CellType", currently_used=True).to_dataframe()
            bt.CellType.import_source()

            # import all records from a specific source
            source = bt.Source.get(entity="bionty.CellType", source="cl", version="2022-08-16")
            bt.CellType.import_source(source)
            bt.CellType.to_dataframe()  # all records from the source are now in the registry

            # update existing records with a new source (version update)
            source = bt.Source.get(entity="bionty.CellType", source="cl", version="2024-08-16")
            bt.CellType.import_source(source, update_records=True)
        """
        if isinstance(source, str):
            raise TypeError("import_source() expects a `bt.Source` object, not a str.")

        if update_records:
            from .core._source import update_records_to_source

            update_records_to_source(cls, source)
        else:
            from .core._add_ontology import add_ontology_from_df

            add_ontology_from_df(
                registry=cls,
                organism=organism,
                source=source,
                ignore_conflicts=ignore_conflicts,
            )

    @classmethod
    def add_source(
        cls,
        source: Source | PublicOntology | str,
        *,
        df: pd.DataFrame | None = None,
        version: str | None = None,
        organism: str | None = None,
    ) -> Source:
        """Link a source record to the entity with a reference DataFrame.

        Creates or retrieves a Source record for the entity and optionally associates
        it with a DataFrame artifact containing the ontology data. If the source
        already exists with a DataFrame artifact, returns the existing source.

        Args:
            source: Source specification. Can be:

                - Source record: Existing :class:`bionty.Source` instance
                - PublicOntology: PublicOntology object with source metadata
                - str: Source name (e.g., "mondo", "cl", "go")

            df: Optional DataFrame containing ontology data to store as Artifact.
                If None and source is a PublicOntology, uses the ontology's DataFrame.
            version: Source version string. Required when source is str and no existing source found.
                Examples: "2025-06-03", "v1.0", "release-112"
            organism: Organism identifier. Required for organism-specific entities when source is str.
                Use "all" for cross-organism ontologies.

        Examples:
            Add a source by name with version and organism::

                import bionty as bt
                source = bt.Disease.add_source("mondo", version="2025-06-03", organism="all")

            Add a source to an entity with a custom DataFrame::

                import pandas as pd
                df = pd.DataFrame({"name": ["disease1"], "ontology_id": ["MONDO:123"]})
                source = bt.Source(
                    entity="bionty.Disease",
                    name="new mondo",
                    version="99.999",
                    organism="human",
                )
                source = bt.Disease.add_source(source=source, df=df)

            Add from existing PublicOntology::

                pub_ont = bt.Disease.public()
                source = bt.Disease.add_source(pub_ont)

            Add organism-specific source::

                source = bt.Gene.add_source("ensembl", version="release-112", organism="human")
        """
        import lamindb as ln

        from bionty.base._public_ontology import encode_filenames

        from ._organism import is_organism_required

        # wetlab.Compound, bionty.CellType, etc.
        entity_name = cls.__get_name_with_module__()
        source_record = source if isinstance(source, Source) else None
        parquet_filename = None

        # Process source input and get source_record
        if isinstance(source, str):
            filter_kwargs = {"name": source, "entity": entity_name}
            if version:
                filter_kwargs["version"] = version
            if organism:
                filter_kwargs["organism"] = organism

            # if source is new and hasn't been registered in the instance yet
            # try to download the source via PublicOntology
            source_record = Source.filter(**filter_kwargs).first()
            if not source_record:
                source = PublicOntology(
                    source=source,
                    version=version,
                    organism=organism,
                    entity=entity_name,
                )
        if source_record is None and isinstance(source, PublicOntology):
            parquet_filename = source._parquet_filename
            source_record = Source(**source._source_dict, entity=entity_name)

        # Core source metadata
        unique_kwargs = {
            "entity": entity_name,
            "name": source_record.name,
            "version": source_record.version,
            "organism": source_record.organism,
        }

        # Register organism if required
        if is_organism_required(cls):
            Organism.from_source(name=source_record.organism).save()

        # Get existing or create new source
        new_source = Source.filter(**unique_kwargs).one_or_none()
        if not new_source:
            new_source = Source(  # type:ignore
                **unique_kwargs,
                currently_used=source_record.currently_used,
                description=source_record.description,
                url=source_record.url,
                source_website=source_record.source_website,
                dataframe_artifact_id=source_record.dataframe_artifact_id,
            ).save()

        # Return early if artifact already exists
        if new_source.dataframe_artifact_id:
            return new_source

        # Generate filename if not already created
        if not parquet_filename:
            parquet_filename, _ = encode_filenames(**unique_kwargs)

        # Create dataframe artifact if needed
        df_artifact = None
        if isinstance(df, pd.DataFrame):
            # backwards compatible
            df_artifact = getattr(ln.Artifact, "from_dataframe", ln.Artifact.from_df)(
                df, key=parquet_filename, run=False
            )
        elif source_record.url and source_record.url.startswith("s3://bionty-assets/"):
            df_artifact = ln.Artifact(new_source.url, run=False)
        elif (
            # for bionty-assets, we do not create dataframe artifact here but with register_source_in_bionty_assets
            ln.setup.settings.instance.slug != "laminlabs/bionty-assets"
            and isinstance(source, PublicOntology)
            and not source.to_dataframe().empty
        ):
            # backwards compatible
            df_artifact = getattr(ln.Artifact, "from_dataframe", ln.Artifact.from_df)(
                source.to_dataframe(), key=parquet_filename, run=False
            )

        # Save dataframe artifact and update source
        if df_artifact:
            df_artifact.kind = "__lamindb_run__"
            df_artifact.save()
            new_source.dataframe_artifact = df_artifact
            new_source.save()

        logger.success(
            f"Added source '{new_source.name}' of version '{new_source.version}' for organism '{new_source.organism}'."
        )
        return new_source

    @classmethod
    def public(
        cls,
        organism: str | SQLRecord | None = None,
        source: Source | None = None,
    ) -> PublicOntology | StaticReference:
        """The corresponding :class:`docs:bionty.base.PublicOntology` object.

        Note that the source is auto-configured and tracked via :class:`docs:bionty.Source`.

        Args:
            organism: Organism name or record to filter by
            source: Source record to use instead of default

        See Also:
            :doc:`docs:public-ontologies`

        Example::

            import bionty as bt

            # default source
            celltype_pub = bt.CellType.public()
            celltype_pub
            #> PublicOntology
            #> Entity: CellType
            #> Organism: all
            #> Source: cl, 2023-04-20
            #> #terms: 2698

            # default source of a organism
            gene_pub = bt.Gene.public(organism="mouse")
            gene_pub
            #> PublicOntology
            #> Entity: Gene
            #> Organism: mouse
            #> Source: ensembl, release-112
            #> #terms: 57510
        """
        if isinstance(organism, Organism):
            organism = organism.name

        if source is not None:
            organism = source.organism
            source_name = source.name
            version = source.version
        else:
            from ._source import get_source_record
            from .core._settings import settings

            if hasattr(cls, "organism_id"):
                if organism is None and settings.organism is not None:
                    organism = settings.organism.name
            source_name = None
            version = None
            source = get_source_record(cls, organism=organism)
            if source is not None:
                source_name = source.name
                version = source.version

        if source is not None and source.dataframe_artifact_id is not None:
            # if source is a DataFrame artifact, return the static reference
            return StaticReference(source)
        if source is None:
            kwargs = {
                "entity": cls.__get_name_with_module__(),
                "currently_used": True,
            }
            if organism is not None:
                if isinstance(organism, SQLRecord):
                    kwargs["organism"] = organism.name
                else:
                    kwargs["organism"] = organism
            source = Source.filter(**kwargs).first()
            if source and source.dataframe_artifact_id is not None:
                return StaticReference(source)

        try:
            return getattr(bt_base, cls.__name__)(
                organism=organism, source=source_name, version=version
            )
        except InvalidParamError as e:
            raise ValueError(str(e)) from None
        except (AttributeError, ValueError):
            if source is None:
                kwargs = {
                    "entity": cls.__get_name_with_module__(),
                    "currently_used": True,
                }
                if organism is not None:
                    if isinstance(organism, SQLRecord):
                        kwargs["organism"] = organism.name
                    else:
                        kwargs["organism"] = organism
                source = Source.filter(**kwargs).first()
            return StaticReference(source)

    @classmethod
    def from_source(
        cls, *, mute: bool = False, **kwargs
    ) -> BioRecord | list[BioRecord] | None:
        """Create a record or records from source based on a single field value.

        Notes:
            For more info, see tutorial :doc:`docs:bionty`

            Bulk create records via :meth:`.from_values`.

        Example::

            import bionty as bt

            # Create a record by passing a field value:
            record = bt.Gene.from_source(symbol="TCF7", organism="human")

            # Create a record from non-default source:
            source = bt.Source.get(entity="CellType", source="cl", version="2022-08-16")
            record = bt.CellType.from_source(name="T cell", source=source)

        """
        # non-relationship kwargs
        kv = {
            k: v
            for k, v in kwargs.items()
            if k not in [i.name for i in cls._meta.fields if i.is_relation]
        }
        if len(kv) > 1:
            raise InvalidArgument(
                "Only one field can be passed to generate records from source"
            )
        elif len(kv) == 0:
            raise InvalidArgument("No field passed to generate records from source")
        else:
            k = next(iter(kv))
            v = kwargs.pop(k)
            results = cls.from_values([v], field=getattr(cls, k), mute=mute, **kwargs)
            if len(results) == 1:
                return results[0]
            elif len(results) == 0:
                raise DoesNotExist(
                    "No record found in source for the given field value"
                )
            else:
                return results

    def save(self, *args, **kwargs) -> BioRecord:
        """Save the record and its parents recursively.

        Example::

            import bionty as bt

            record = bt.CellType.from_source(name="T cell")
            record.save()
        """
        super().save(*args, **kwargs)
        # saving records of parents
        if hasattr(self, "_parents"):
            import lamindb as ln

            # here parents is still a list of ontology ids
            parents = self._parents
            # bulk create parent records
            parents_records = self.from_values(
                parents, self.__class__.ontology_id, source=self.source
            )
            ln.save(parents_records)
            self.parents.set(parents_records)

        return self


class Organism(BioRecord, TracksRun, TracksUpdates):
    """Organism - `NCBI Taxonomy <https://www.ncbi.nlm.nih.gov/taxonomy/>`__, `Ensembl Organism <https://useast.ensembl.org/info/about/species.html>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:organism`.

    Example::

        import bionty as bt

        record = bt.Organism.from_source(name="rabbit")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=64, db_index=True, default=None, unique=True)
    """Name of a organism, required field."""
    ontology_id: str | None = CharField(
        max_length=32, unique=True, db_index=True, null=True, default=None
    )
    """NCBI Taxon ID."""
    scientific_name: str | None = CharField(
        max_length=64, db_index=True, unique=True, null=True, default=None
    )
    """Scientific name of a organism."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this organism."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the organism."""
    parents: Organism = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent organism records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactOrganism", related_name="organisms"
    )
    """Artifacts linked to the organism."""
    records: Record = models.ManyToManyField(
        Record, through="RecordOrganism", related_name="organisms"
    )
    """Records linked to the organism."""

    @overload
    def __init__(
        self,
        name: str,
        taxon_id: str | None,
        scientific_name: str | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        scientific_name: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Organism | list[Organism] | None:
        """Create an Organism record from source based on a single identifying field.

        Args:
            name: Common name of the organism (e.g. "human", "mouse", "zebrafish")
            ontology_id: NCBI Taxon ID (e.g. "9606")
            scientific_name: Scientific name (e.g. "Homo sapiens", "Mus musculus")
            {doc_from_source}

        Returns:
            A single Organism record, list of Organism records, or None if not found

        Example::

            import bionty as bt

            record = bt.Organism.from_source(name="human")
            record = bt.Organism.from_source(ontology_id="NCBITaxon:9606")
        """
        pass


class Gene(BioRecord, TracksRun, TracksUpdates):
    """Genes - `Ensembl <https://ensembl.org/>`__, `NCBI Gene <https://www.ncbi.nlm.nih.gov/gene/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:gene`.

        Bulk create Gene records via `.from_values()`.
        Map legacy ensembl IDs to current ensembl IDs using :meth:`bionty.base.Gene.map_legacy_ids`.

        We discourage validating gene symbols and to work with unique identifiers such as ENSEMBL IDs instead.
        For more details, see :doc:`docs:faq/symbol-mapping`.

    Example::

        import bionty as bt

        record = bt.Gene.from_source(ensembl_gene_id="ENSG00000081059")
        record = bt.Gene.from_source(symbol="TCF7", organism="human")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"

    _name_field: str = "symbol"
    _ontology_id_field: str = "ensembl_gene_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=12, db_index=True, default=ids.gene)
    """A universal id (base62-encoded hash of defining fields)."""
    symbol: str | None = CharField(
        max_length=64, db_index=True, null=True, default=None
    )
    """A unique short form of gene name."""
    stable_id: str | None = CharField(
        max_length=64, db_index=True, null=True, default=None, unique=True
    )
    """Stable ID of a gene that doesn't have ensembl_gene_id, e.g. a yeast gene."""
    ensembl_gene_id: str | None = CharField(
        max_length=64, db_index=True, null=True, default=None, unique=True
    )
    """Ensembl gene stable ID, in the form ENS[organism prefix][feature type prefix][a unique eleven digit number]."""
    ncbi_gene_ids: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) NCBI Gene IDs that correspond to this Ensembl Gene ID.
    NCBI Gene ID, also known as Entrez Gene ID, in the form of numeric string, 1 to 9 digits.
    """
    biotype: str | None = CharField(
        max_length=64, db_index=True, null=True, default=None
    )
    """Type of the gene."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this gene."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the gene."""
    organism: Organism = ForeignKey(
        Organism, PROTECT, default=None, related_name="genes"
    )
    """:class:`~bionty.Organism` of the gene."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactGene", related_name="genes"
    )
    """Artifacts linked to the gene."""
    records: Record = models.ManyToManyField(
        Record, through="RecordGene", related_name="genes"
    )
    """Records linked to the gene."""
    schemas: Schema = models.ManyToManyField(
        Schema, through="SchemaGene", related_name="genes"
    )
    """Schemas linked to this gene."""

    @overload
    def __init__(
        self,
        symbol: str | None,
        stable_id: str | None,
        ensembl_gene_id: str | None,
        ncbi_gene_ids: str | None,
        biotype: str | None,
        description: str | None,
        synonyms: str | None,
        organism: Organism | None,
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        symbol: str | None = None,
        ensembl_gene_id: str | None = None,
        stable_id: str | None = None,
        organism: str | Organism = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Gene | list[Gene] | None:
        """Create a Gene record from source based on a single identifying field.

        Args:
            symbol: Gene symbol (e.g. "TCF7")
            ensembl_gene_id: Ensembl gene ID (e.g. "ENSG00000081059")
            stable_id: Stable ID for genes without Ensembl IDs (e.g. yeast genes)
            organism: Organism name or Organism record
            {doc_from_source}

        Returns:
            A single Gene record, list of Gene records, or None if not found

        Example::

            import bionty as bt

            record = bt.Gene.from_source(symbol="TCF7", organism="human")
            record = bt.Gene.from_source(ensembl_gene_id="ENSG00000081059")
            record = bt.Gene.from_source(stable_id="YAL001C", organism="yeast")
        """
        pass


class Protein(BioRecord, TracksRun, TracksUpdates):
    """Proteins - `Uniprot <https://www.uniprot.org/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:protein`.

        Bulk create records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.Protein.from_source(name="Synaptotagmin-15B", organism="human")
        record = bt.Protein.from_source(gene_symbol="SYT15B", organism="human")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"

    _name_field: str = "name"
    _ontology_id_field: str = "uniprotkb_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=12, db_index=True, default=ids.protein)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str | None = CharField(max_length=256, db_index=True, null=True, default=None)
    """Unique name of a protein."""
    uniprotkb_id: str | None = CharField(
        max_length=10, db_index=True, null=True, default=None, unique=True
    )
    """UniProt protein ID, 6 alphanumeric characters, possibly suffixed by 4 more."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this protein."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the protein."""
    length: int | None = BigIntegerField(db_index=True, null=True)
    """Length of the protein sequence."""
    gene_symbol: str | None = CharField(
        max_length=256, db_index=True, null=True, default=None
    )
    """The primary gene symbol corresponds to this protein."""
    ensembl_gene_ids: str | None = TextField(null=True, default=None)
    """Bar-separated (|) Ensembl Gene IDs that correspond to this protein."""
    organism: Organism = ForeignKey(
        Organism, PROTECT, default=None, related_name="proteins"
    )
    """:class:`~bionty.Organism` this protein associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactProtein", related_name="proteins"
    )
    """Artifacts linked to the protein."""
    records: Record = models.ManyToManyField(
        Record, through="RecordProtein", related_name="proteins"
    )
    """Records linked to the protein."""
    schemas: Schema = models.ManyToManyField(
        Schema, through="SchemaProtein", related_name="proteins"
    )
    """Featuresets linked to this protein."""

    @overload
    def __init__(
        self,
        name: str | None,
        uniprotkb_id: str | None,
        synonyms: str | None,
        length: int | None,
        gene_symbol: str | None,
        ensembl_gene_ids: str | None,
        organism: Organism | None,
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        uniprotkb_id: str | None = None,
        gene_symbol: str | None = None,
        organism: str | Organism | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Protein | list[Protein] | None:
        """Create a Protein record from source based on a single identifying field.

        Args:
            name: Protein name (e.g. "Synaptotagmin-15B")
            uniprotkb_id: UniProt protein ID (e.g. "Q8N6N3")
            gene_symbol: Gene symbol (e.g. "SYT15B")
            organism: Organism name or Organism record
            {doc_from_source}

        Returns:
            A single Protein record, list of Protein records, or None if not found

        Example::

            import bionty as bt

            record = bt.Protein.from_source(name="Synaptotagmin-15B", organism="human")
            record = bt.Protein.from_source(uniprotkb_id="Q8N6N3")
            record = bt.Protein.from_source(gene_symbol="SYT15B", organism="human")
        """
        pass


class CellMarker(BioRecord, TracksRun, TracksUpdates):
    """Cell markers - `CellMarker <http://xteam.xbio.top/CellMarker>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:cell_marker`.

        Bulk create CellMarker records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.CellMarker.from_source(name="PD1", organism="human")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "organism"),)

    _name_field: str = "name"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        unique=True, max_length=12, db_index=True, default=ids.cellmarker
    )
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=64, db_index=True)
    """Unique name of the cell marker."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this cell marker."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the cell marker."""
    gene_symbol: str | None = CharField(
        max_length=64, db_index=True, null=True, default=None
    )
    """Gene symbol that corresponds to this cell marker."""
    ncbi_gene_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """NCBI gene id that corresponds to this cell marker."""
    uniprotkb_id: str | None = CharField(
        max_length=10, db_index=True, null=True, default=None
    )
    """Uniprotkb id that corresponds to this cell marker."""
    organism: Organism = ForeignKey(
        Organism, PROTECT, default=None, related_name="cell_markers"
    )
    """:class:`~bionty.Organism` this cell marker associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactCellMarker",
        related_name="cell_markers",
    )
    """Artifacts linked to the cell marker."""
    records: Record = models.ManyToManyField(
        Record,
        through="RecordCellMarker",
        related_name="cell_markers",
    )
    """Records linked to the cell marker."""
    schemas: Schema = models.ManyToManyField(
        Schema, through="SchemaCellMarker", related_name="cell_markers"
    )
    """Featuresets linked to this cell marker."""

    @overload
    def __init__(
        self,
        name: str,
        synonyms: str | None,
        gene_symbol: str | None,
        ncbi_gene_id: str | None,
        uniprotkb_id: str | None,
        organism: Organism | None,
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        gene_symbol: str | None = None,
        ncbi_gene_id: str | None = None,
        uniprotkb_id: str | None = None,
        organism: str | Organism | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> CellMarker | list[CellMarker] | None:
        """Create a CellMarker record from source based on a single identifying field.

        Args:
            name: Cell marker name (e.g. "PD1", "CD19")
            gene_symbol: Gene symbol (e.g. "PDCD1", "CD19")
            ncbi_gene_id: NCBI gene ID that corresponds to this cell marker
            uniprotkb_id: UniProt ID that corresponds to this cell marker
            organism: Organism name or Organism record
            {doc_from_source}

        Returns:
            A single CellMarker record, list of CellMarker records, or None if not found

        Example::

            import bionty as bt

            record = bt.CellMarker.from_source(name="PD1", organism="human")
            record = bt.CellMarker.from_source(gene_symbol="PDCD1", organism="human")
            record = bt.CellMarker.from_source(name="CD19", organism="mouse")
        """
        pass


class Tissue(BioRecord, TracksRun, TracksUpdates):
    """Tissues - `Uberon <http://obophenotype.github.io/uberon/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` :doc:`docs:tissue`.

        Bulk create Tissue records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.Tissue.from_source(name="brain")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the tissue."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the tissue."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of tissue."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this tissue."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the tissue."""
    parents: Tissue = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent tissues records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactTissue", related_name="tissues"
    )
    """Artifacts linked to the tissue."""
    records: Record = models.ManyToManyField(
        Record, through="RecordTissue", related_name="tissues"
    )
    """Records linked to the tissue."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Tissue],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        organism: str | Organism | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Tissue | list[Tissue] | None:
        """Create a Tissue record from source based on a single identifying field.

        Args:
            name: Tissue name (e.g. "nose")
            ontology_id: Tissue ontology ID (e.g. "UBERON:0000004")
            organism: Organism name or Organism record
            {doc_from_source}

        Returns:
            A single Tissue record, list of Tissue records, or None if not found

        Example::

            import bionty as bt

            record = bt.Tissue.from_source(name="nose")
            record = bt.Tissue.from_source(ontology_id="UBERON:0000004")
        """
        pass


class CellType(BioRecord, TracksRun, TracksUpdates):
    """Cell types - `Cell Ontology <https://obophenotype.github.io/cell-ontology/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:cell_type`.

        Bulk create CellType records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.CellType.from_source(name="T cell")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the cell type."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the cell type."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of cell type."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this cell type."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the cell type."""
    parents: CellType = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent cell type records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactCellType", related_name="cell_types"
    )
    """Artifacts linked to the cell type."""
    records: Record = models.ManyToManyField(
        Record, through="RecordCellType", related_name="cell_types"
    )
    """Records linked to the cell type."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[CellType],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> CellType | list[CellType] | None:
        """Create a CellType record from source based on a single identifying field.

        Args:
            name: Name of the cell type (e.g. "T cell", "B cell")
            ontology_id: Cell Ontology ID (e.g. "CL:0000084")
            abbr: Unique abbreviation of cell type
            {doc_from_source}

        Returns:
            A single CellType record, list of CellType records, or None if not found

        Example::

            import bionty as bt

            record = bt.CellType.from_source(name="T cell")
            record = bt.CellType.from_source(ontology_id="CL:0000084")

            source = bt.Source.get(entity="bionty.CellType", source="cl", version="2024-08-16")
            record = bt.CellType.from_source(name="B cell", source=source)
        """
        pass


class Disease(BioRecord, TracksRun, TracksUpdates):
    """Diseases - `Mondo <https://mondo.monarchinitiative.org/>`__, `Human Disease <https://disease-ontology.org/>`__.

    Notes:
        Bulk create Disease records via :meth:`.from_values`.

        For more info, see tutorials: :doc:`docs:disease`.

    Example::

        import bionty as bt

        record = bt.Disease.from_source(name="Alzheimer disease")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the disease."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the disease."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of disease."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this disease."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the disease."""
    parents: Disease = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent disease records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactDisease", related_name="diseases"
    )
    """Artifacts linked to the disease."""
    records: Record = models.ManyToManyField(
        Record, through="RecordDisease", related_name="diseases"
    )
    """Records linked to the disease."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Disease],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Disease | list[Disease] | None:
        """Create a Disease record from source based on a single identifying field.

        Args:
            name: Name of the disease (e.g. "Alzheimer disease", "type 2 diabetes")
            ontology_id: Disease ontology ID (e.g. "MONDO:0004975")
            abbr: Unique abbreviation of disease
            {doc_from_source}

        Returns:
            A single Disease record, list of Disease records, or None if not found

        Example::

            import bionty as bt

            record = bt.Disease.from_source(name="Alzheimer disease")
            record = bt.Disease.from_source(ontology_id="MONDO:0004975")
            record = bt.Disease.from_source(name="type 2 diabetes")
        """
        pass


class CellLine(BioRecord, TracksRun, TracksUpdates):
    """Cell lines - `Cell Line Ontology <https://github.com/CLO-ontology/CLO>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:cell_line`.

        Bulk create CellLine records via :meth:`.from_values`.

    Example::

        import bionty as bt

        standard_name = bt.CellLine.public().standardize(["K562"])[0]
        record = bt.CellLine.from_source(name=standard_name)
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the cell line."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the cell line."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of cell line."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this cell line."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the cell line."""
    parents: CellLine = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent cell line records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactCellLine", related_name="cell_lines"
    )
    """Artifacts linked to the cell line."""
    records: Record = models.ManyToManyField(
        Record, through="RecordCellLine", related_name="cell_lines"
    )
    """Records linked to the cell line."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[CellLine],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> CellLine | list[CellLine] | None:
        """Create a CellLine record from source based on a single identifying field.

        Args:
            name: Name of the cell line (e.g. "K562", "HeLa")
            ontology_id: Cell Line Ontology ID (e.g. "CLO:0009477")
            abbr: Unique abbreviation of cell line
            {doc_from_source}

        Returns:
            A single CellLine record, list of CellLine records, or None if not found

        Example::

            import bionty as bt

            record = bt.CellLine.from_source(name="K562")
            record = bt.CellLine.from_source(ontology_id="CLO:0009477")
        """
        pass


class Phenotype(BioRecord, TracksRun, TracksUpdates):
    """Phenotypes - `Human Phenotype <https://hpo.jax.org/app/>`__,
    `Phecodes <https://phewascatalog.org/phecodes_icd10>`__,
    `Mammalian Phenotype <http://obofoundry.org/ontology/mp.html>`__,
    `Zebrafish Phenotype <http://obofoundry.org/ontology/zp.html>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:phenotype`.

        Bulk create Phenotype records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.Phenotype.from_source(name="Arachnodactyly")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the phenotype."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the phenotype."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of phenotype."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this phenotype."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the phenotype."""
    parents: Phenotype = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent phenotype records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactPhenotype", related_name="phenotypes"
    )
    """Artifacts linked to the phenotype."""
    records: Record = models.ManyToManyField(
        Record, through="RecordPhenotype", related_name="phenotypes"
    )
    """Records linked to the phenotype."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Phenotype],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Phenotype | list[Phenotype] | None:
        """Create a Phenotype record from source based on a single identifying field.

        Args:
            name: Name of the phenotype (e.g. "Arachnodactyly", "Cardiomegaly")
            ontology_id: Phenotype ontology ID (e.g. "HP:0001166")
            abbr: Unique abbreviation of phenotype
            {doc_from_source}

        Returns:
            A single Phenotype record, list of Phenotype records, or None if not found

        Example::

            import bionty as bt

            record = bt.Phenotype.from_source(name="Arachnodactyly")
            record = bt.Phenotype.from_source(ontology_id="HP:0001166")
        """
        pass


class Pathway(BioRecord, TracksRun, TracksUpdates):
    """Pathways - `Gene Ontology <https://bioportal.bioontology.org/ontologies/GO>`__,
    `Pathway Ontology <https://bioportal.bioontology.org/ontologies/PW>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:pathway`.

        Bulk create Pathway records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.Pathway.from_source(ontology_id="GO:1903353")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the pathway."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the pathway."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of pathway."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this pathway."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the pathway."""
    parents: Pathway = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent pathway records."""
    genes: Gene = models.ManyToManyField("Gene", related_name="pathways")
    """Genes that signifies the pathway."""
    schemas: Schema = models.ManyToManyField(
        Schema, through="SchemaPathway", related_name="pathways"
    )
    """Featuresets linked to the pathway."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactPathway", related_name="pathways"
    )
    """Artifacts linked to the pathway."""
    records: Record = models.ManyToManyField(
        Record, through="RecordPathway", related_name="pathways"
    )
    """Records linked to the pathway."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Pathway],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Pathway | list[Pathway] | None:
        """Create a Pathway record from source based on a single identifying field.

        Args:
            name: Name of the pathway (e.g. "mitotic cell cycle", "glycolysis")
            ontology_id: GO or Pathway Ontology ID (e.g. "GO:1903353", "PW:0000004")
            abbr: Unique abbreviation of pathway
            {doc_from_source}

        Returns:
            A single Pathway record, list of Pathway records, or None if not found

        Example::

            import bionty as bt

            record = bt.Pathway.from_source(name="mitotic cell cycle")
            record = bt.Pathway.from_source(ontology_id="GO:1903353")
        """
        pass


class ExperimentalFactor(BioRecord, TracksRun, TracksUpdates):
    """Experimental factors - `Experimental Factor Ontology <https://www.ebi.ac.uk/ols/ontologies/efo>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:experimental_factor`.

        Bulk create ExperimentalFactor records via :meth:`.from_values`.

    Example::

        import bionty as bt

        standard_name = bt.ExperimentalFactor.public().standardize(["scRNA-seq"])
        record = bt.ExperimentalFactor.from_source(name=standard_name)
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the experimental factor."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the experimental factor."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of experimental factor."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this experimental factor."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the experimental factor."""
    molecule: str | None = TextField(null=True, default=None, db_index=True)
    """Molecular experimental factor, parsed from EFO."""
    instrument: str | None = TextField(null=True, default=None, db_index=True)
    """Instrument used to measure the experimental factor, parsed from EFO."""
    measurement: str | None = TextField(null=True, default=None, db_index=True)
    """Phenotypic experimental factor, parsed from EFO."""
    parents: ExperimentalFactor = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent experimental factor records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactExperimentalFactor",
        related_name="experimental_factors",
    )
    """Artifacts linked to the experimental_factors."""
    records: Record = models.ManyToManyField(
        Record,
        through="RecordExperimentalFactor",
        related_name="experimental_factors",
    )
    """Records linked to the experimental_factors."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[ExperimentalFactor],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> ExperimentalFactor | list[ExperimentalFactor] | None:
        """Create an ExperimentalFactor record from source based on a single identifying field.

        Args:
            name: Name of the experimental factor (e.g. "scRNA-seq", "ChIP-seq")
            ontology_id: Experimental Factor Ontology ID (e.g. "EFO:0009922")
            abbr: Unique abbreviation of experimental factor
            {doc_from_source}

        Returns:
            A single ExperimentalFactor record, list of ExperimentalFactor records, or None if not found

        Example::

            import bionty as bt

            record = bt.ExperimentalFactor.from_source(name="scRNA-seq")
            record = bt.ExperimentalFactor.from_source(ontology_id="EFO:0009922")
        """
        pass


class DevelopmentalStage(BioRecord, TracksRun, TracksUpdates):
    """Developmental stages - `Human Developmental Stages <https://github.com/obophenotype/developmental-stage-ontologies/wiki/HsapDv>`__,
    `Mouse Developmental Stages <https://github.com/obophenotype/developmental-stage-ontologies/wiki/MmusDv>`__.  # noqa.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:developmental_stage`.

        Bulk create DevelopmentalStage records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.DevelopmentalStage.from_source(name="neurula stage")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the developmental stage."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the developmental stage."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of developmental stage."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this developmental stage."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the developmental stage."""
    parents: str | None = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent developmental stage records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactDevelopmentalStage",
        related_name="developmental_stages",
    )
    """Artifacts linked to the developmental stage."""
    records: Record = models.ManyToManyField(
        Record,
        through="RecordDevelopmentalStage",
        related_name="developmental_stages",
    )
    """Records linked to the developmental stage."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[DevelopmentalStage],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> DevelopmentalStage | list[DevelopmentalStage] | None:
        """Create a DevelopmentalStage record from source based on a single identifying field.

        Args:
            name: Name of the developmental stage (e.g. "neurula stage", "gastrula stage")
            ontology_id: Developmental stage ontology ID (e.g. "HsapDv:0000004")
            abbr: Unique abbreviation of developmental stage
            {doc_from_source}

        Returns:
            A single DevelopmentalStage record, list of DevelopmentalStage records, or None if not found

        Example::

            import bionty as bt

            record = bt.DevelopmentalStage.from_source(name="neurula stage")
            record = bt.DevelopmentalStage.from_source(ontology_id="HsapDv:0000004")
        """
        pass


class Ethnicity(BioRecord, TracksRun, TracksUpdates):
    """Ethnicity - `Human Ancestry Ontology <https://github.com/EBISPOT/hancestro>`__.

    Notes:
        For more info, see tutorials :doc:`docs:manage-ontologies` and :doc:`docs:ethnicity`.

        Bulk create Ethnicity records via :meth:`.from_values`.

    Example::

        import bionty as bt

        record = bt.Ethnicity.from_source(name="European")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "bionty"
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(unique=True, max_length=8, db_index=True, default=ids.ontology)
    """A universal id (base62-encoded hash of defining fields)."""
    name: str = CharField(max_length=256, db_index=True)
    """Name of the ethnicity."""
    ontology_id: str | None = CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the ethnicity."""
    abbr: str | None = CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of ethnicity."""
    synonyms: str | None = TextField(null=True, db_index=True, default=None)
    """Bar-separated (|) synonyms that correspond to this ethnicity."""
    description: str | None = TextField(null=True, db_index=True, default=None)
    """Description of the ethnicity."""
    parents: Ethnicity = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent ethnicity records."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactEthnicity",
        related_name="ethnicities",
    )
    """Artifacts linked to the ethnicity."""
    records: Record = models.ManyToManyField(
        Record,
        through="RecordEthnicity",
        related_name="ethnicities",
    )
    """Records linked to the ethnicity."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Ethnicity],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    @classmethod
    @pass_to_super
    @_doc_params(doc_from_source=doc_from_source)
    def from_source(
        cls,
        *,
        name: str | None = None,
        ontology_id: str | None = None,
        abbr: str | None = None,
        source: Source | None = None,
        mute: bool = False,
        **kwargs,
    ) -> Ethnicity | list[Ethnicity] | None:
        """Create an Ethnicity record from source based on a single identifying field.

        Args:
            name: Name of the ethnicity (e.g. "European", "East Asian")
            ontology_id: Human Ancestry Ontology ID (e.g. "HANCESTRO:0005")
            abbr: Unique abbreviation of ethnicity
            {doc_from_source}

        Returns:
            A single Ethnicity record, list of Ethnicity records, or None if not found

        Example::

            import bionty as bt

            record = bt.Ethnicity.from_source(name="European")
            record = bt.Ethnicity.from_source(ontology_id="HANCESTRO:0005")
        """
        pass


class SchemaGene(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    schema: Schema = ForeignKey("lamindb.Schema", CASCADE, related_name="links_gene")
    gene: Gene = ForeignKey("Gene", PROTECT, related_name="links_schema")

    class Meta:
        app_label = "bionty"
        unique_together = ("schema", "gene")


class SchemaProtein(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    schema: Schema = ForeignKey("lamindb.Schema", CASCADE, related_name="links_protein")
    protein: Protein = ForeignKey("Protein", PROTECT, related_name="links_schema")

    class Meta:
        app_label = "bionty"
        unique_together = ("schema", "protein")


class SchemaCellMarker(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    schema: Schema = ForeignKey(
        "lamindb.Schema", CASCADE, related_name="links_cellmarker"
    )
    cellmarker: CellMarker = ForeignKey(
        "CellMarker", PROTECT, related_name="links_schema"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("schema", "cellmarker")


class SchemaPathway(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    schema: Schema = ForeignKey("lamindb.Schema", CASCADE, related_name="links_pathway")
    pathway: Pathway = ForeignKey("Pathway", PROTECT, related_name="links_schema")

    class Meta:
        app_label = "bionty"
        unique_together = ("schema", "pathway")


class ArtifactOrganism(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_organism")
    organism: Organism = ForeignKey("Organism", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactorganism"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "organism", "feature")


class ArtifactGene(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_gene")
    gene: Gene = ForeignKey("Gene", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactgene"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "gene", "feature")


class ArtifactProtein(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_protein")
    protein: Protein = ForeignKey("Protein", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactprotein"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "protein", "feature")


class ArtifactCellMarker(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_cell_marker")
    # follow the .lower() convention in link models
    cellmarker: CellMarker = ForeignKey(
        "CellMarker", PROTECT, related_name="links_artifact"
    )
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactcellmarker",
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "cellmarker", "feature")


class ArtifactTissue(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_tissue")
    tissue: Tissue = ForeignKey("Tissue", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifacttissue"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "tissue", "feature")


class ArtifactCellType(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_cell_type")
    # follow the .lower() convention in link models
    celltype: CellType = ForeignKey("CellType", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactcelltype"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "celltype", "feature")


class ArtifactDisease(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_disease")
    disease: Disease = ForeignKey("Disease", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactdisease"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "disease", "feature")


class ArtifactCellLine(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_cell_line")
    # follow the .lower() convention in link models
    cellline: CellLine = ForeignKey("CellLine", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactcellline"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "cellline", "feature")


class ArtifactPhenotype(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_phenotype")
    phenotype: Phenotype = ForeignKey(
        "Phenotype", PROTECT, related_name="links_artifact"
    )
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactphenotype",
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "phenotype", "feature")


class ArtifactPathway(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_pathway")
    pathway: Pathway = ForeignKey("Pathway", PROTECT, related_name="links_artifact")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactpathway"
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "pathway", "feature")


class ArtifactExperimentalFactor(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(
        Artifact, CASCADE, related_name="links_experimental_factor"
    )
    experimentalfactor: ExperimentalFactor = ForeignKey(
        "ExperimentalFactor", PROTECT, related_name="links_artifact"
    )
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactexperimentalfactor",
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "experimentalfactor", "feature")


class ArtifactDevelopmentalStage(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(
        Artifact, CASCADE, related_name="links_developmental_stage"
    )
    # follow the .lower() convention in link models
    developmentalstage: DevelopmentalStage = ForeignKey(
        "DevelopmentalStage", PROTECT, related_name="links_artifact"
    )
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactdevelopmentalstage",
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "developmentalstage", "feature")


class ArtifactEthnicity(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_ethnicity")
    ethnicity: Ethnicity = ForeignKey(
        "Ethnicity", PROTECT, related_name="links_artifact"
    )
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactethnicity",
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "bionty"
        unique_together = ("artifact", "ethnicity", "feature")


class RecordOrganism(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_organism")
    value: Organism = ForeignKey("Organism", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recordorganism"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordGene(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_gene")
    value: Gene = ForeignKey("Gene", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recordgene"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordProtein(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_protein")
    value: Protein = ForeignKey("Protein", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recordprotein"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordCellMarker(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_cell_marker")
    value: CellMarker = ForeignKey("CellMarker", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_recordcellmarker",
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordTissue(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_tissue")
    value: Tissue = ForeignKey("Tissue", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recordtissue"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordCellType(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_cell_type")
    value: CellType = ForeignKey("CellType", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recordcelltype"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordDisease(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_disease")
    value: Disease = ForeignKey("Disease", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recorddisease"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordCellLine(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_cell_line")
    value: CellLine = ForeignKey("CellLine", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recordcellline"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordPhenotype(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_phenotype")
    value: Phenotype = ForeignKey("Phenotype", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_recordphenotype",
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordPathway(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_pathway")
    value: Pathway = ForeignKey("Pathway", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_recordpathway"
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordExperimentalFactor(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(
        Record, CASCADE, related_name="values_experimental_factor"
    )
    value: ExperimentalFactor = ForeignKey(
        "ExperimentalFactor", PROTECT, related_name="links_record"
    )
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_recordexperimentalfactor",
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordDevelopmentalStage(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(
        Record, CASCADE, related_name="values_developmental_stage"
    )
    value: DevelopmentalStage = ForeignKey(
        "DevelopmentalStage", PROTECT, related_name="links_record"
    )
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_recorddevelopmentalstage",
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


class RecordEthnicity(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_ethnicity")
    value: Ethnicity = ForeignKey("Ethnicity", PROTECT, related_name="links_record")
    feature: Feature = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_recordethnicity",
    )

    class Meta:
        app_label = "bionty"
        unique_together = ("record", "value", "feature")


# backward compat
Species = Organism
BiontySource = Source
BioRegistry = BioRecord
PublicSource = Source



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from bionty.base._public_ontology import PublicOntology
from bionty.base.dev._doc_util import _doc_params
from bionty.base.entities._shared_docstrings import organism_removed

if TYPE_CHECKING:
    from bionty.base._ontology import Ontology


@_doc_params(doc_entities=organism_removed)
class ExperimentalFactor(PublicOntology):
    """Experimental Factor.

    1. Experimental Factor Ontology
    https://www.ebi.ac.uk/ols/ontologies/efo

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Literal["all"] | None = None,
        source: Literal["efo"] | None = None,
        version: Literal[
            "3.78.0",
            "3.70.0",
            "3.69.0",
            "3.66.0",
            "3.65.0",
            "3.63.0",
            "3.62.0",
            "3.57.0",
            "3.48.0",
        ]
        | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            organism=organism,
            source=source,
            version=version,
            include_id_prefixes={"efo": ["EFO:", "http://www.ebi.ac.uk/efo/"]},
            **kwargs,
        )

    def to_pronto(self, mute: bool = False) -> Ontology:
        """The Pronto Ontology object.

        See: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html
        """
        from bionty.base._ontology import Ontology

        self._download_ontology_file(
            localpath=self._local_ontology_path,
            url=self._url,
        )
        return Ontology(
            handle=self._local_ontology_path,
            prefix="http://www.ebi.ac.uk/efo/",
        )



from pathlib import Path

from bionty.base._ontology import Ontology
from bionty.base.dev._io import s3_bionty_assets


def test_ontology():
    localpath = s3_bionty_assets("ontology_all__pw__7.79__Pathway")

    try:
        onto = Ontology(localpath)
        assert onto.get_term("PW:0000014").name == "neurodegenerative pathway"

        df = onto.to_df(source="pw", include_id_prefixes={"pw": ["PW"]})
        assert df.shape == (2647, 4)
        assert df.index.name == "ontology_id"

    finally:
        if Path(localpath).exists:
            Path(localpath).unlink()



from __future__ import annotations

from typing import Literal

from bionty.base._public_ontology import PublicOntology
from bionty.base.dev._doc_util import _doc_params

from ._shared_docstrings import doc_entites


@_doc_params(doc_entities=doc_entites)
class BFXPipeline(PublicOntology):
    """Bioinformatics pipelines.

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Literal["all"] | None = None,
        source: Literal["lamin"] | None = None,
        version: Literal["1.0.0"] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            ols_supported=False,
            **kwargs,
        )



# Generated by Django 5.2 on 2025-07-05 15:36

import django.db.models.deletion
import lamindb.base.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0055_rename_cellline_recordcellline_value_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="recordtissue",
            name="record",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="values_tissue",
                to="lamindb.record",
            ),
        ),
    ]



# Generated by Django 5.2 on 2025-05-25 12:04

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0051_alter_cellline__branch_code_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="cellline",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="cellmarker",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="celltype",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="developmentalstage",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="disease",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="ethnicity",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="experimentalfactor",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="gene",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="organism",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="pathway",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="phenotype",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="protein",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="source",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="tissue",
            old_name="_branch_code",
            new_name="branch",
        ),
    ]



from functools import lru_cache

import requests


def import_bioregistry():
    """Import bioregistry module if available."""
    try:
        import bioregistry

        return bioregistry
    except ImportError:
        raise ImportError(
            "Please install bioregistry with `pip install bioregistry`."
        ) from None


class OntologyURLError(Exception):
    """Base exception for ontology resolver errors."""

    pass


class OntologyNotFoundError(OntologyURLError):
    """Raised when an ontology cannot be found."""

    pass


class OntologyVersionNotFoundError(OntologyURLError):
    """Raised when a specific version of an ontology cannot be found."""

    pass


@lru_cache(maxsize=128)
def get_ontology_url(prefix: str, version: str | None = None) -> tuple[str, str]:
    """Get a versioned download URL for an ontology based on its CURIE prefix.

    Args:
        prefix: The CURIE prefix (e.g., 'GO', 'MONDO', 'HP')
        version: Optional version string (e.g., '2023-01-01')
                If None, the latest version will be determined.

    Returns:
        Tuple of (download_url, version_string)

    Raises:
        OntologyNotFoundError: If the ontology cannot be found
        OntologyVersionNotFoundError: If no versioned URL can be found
    """
    bioregistry = import_bioregistry()

    if not prefix:
        raise ValueError("please provide a prefix")

    # Normalize the prefix
    normalized = bioregistry.normalize_prefix(prefix) or prefix

    # Check if the prefix exists at all
    if not _prefix_exists(normalized):
        raise OntologyNotFoundError(f"ontology with prefix '{prefix}' not found")

    # If specific version requested, try to get it
    if version:
        # try standard versioned URL patterns
        url, ver = _get_specific_version(normalized, version)
        if url:
            return url, ver

        raise OntologyVersionNotFoundError(
            f"version '{version}' of ontology '{prefix}' not found"
        )

    # For latest version
    url, ver = _get_latest_from_ols4(normalized)
    if url:
        return url, ver

    # If we get here, no versioned URL was found
    raise OntologyVersionNotFoundError(
        f"no versioned URL found for ontology '{prefix}'"
    )


def _prefix_exists(prefix: str) -> bool:
    """Check if a prefix exists in any registry."""
    bioregistry = import_bioregistry()

    if bioregistry.normalize_prefix(prefix):
        return True

    # Check OLS4
    try:
        response = requests.head(
            f"https://www.ebi.ac.uk/ols4/api/ontologies/{prefix.lower()}", timeout=5
        )
        if response.status_code < 400:
            return True
    except requests.RequestException:
        pass

    return False


def _url_exists(url: str) -> bool:
    """Check if a URL exists and returns a valid response."""
    try:
        response = requests.head(url, timeout=5, allow_redirects=True)
        return response.status_code >= 200 and response.status_code < 400
    except requests.RequestException:
        return False


def _extract_version_from_iri(version_iri: str | None):
    """Extract version from an IRI string by taking the second-to-last path component."""
    if isinstance(version_iri, str):
        # If we have at least two parts, return the second-to-last
        parts = version_iri.split("/")
        if len(parts) >= 2:
            return parts[-2].removeprefix("v")


def _get_specific_version(prefix: str, version: str) -> tuple[str | None, str | None]:
    """Get URL for a specific version of an ontology using standard patterns."""
    bioregistry = import_bioregistry()

    # Clean version string
    clean_version = version[1:] if version.startswith("v") else version
    obo_prefix = bioregistry.get_obofoundry_prefix(prefix) or prefix

    # Try standard OBO Foundry versioned patterns
    standard_patterns = [
        # Direct version path
        f"http://purl.obolibrary.org/obo/{obo_prefix.lower()}/{clean_version}/{obo_prefix.lower()}.owl",
        # Releases directory
        f"http://purl.obolibrary.org/obo/{obo_prefix.lower()}/releases/{clean_version}/{obo_prefix.lower()}.owl",
        # Semantic version with v prefix
        f"http://purl.obolibrary.org/obo/{obo_prefix.lower()}/v{clean_version}/{obo_prefix.lower()}.owl",
    ]

    for url in standard_patterns:
        if _url_exists(url):
            return url, clean_version

    return None, None


def _get_latest_from_ols4(prefix: str) -> tuple[str | None, str | None]:
    """Get the latest version information from OLS4."""
    try:
        response = requests.get(
            f"https://www.ebi.ac.uk/ols4/api/ontologies/{prefix.lower()}", timeout=30
        )
        if response.status_code != 200:
            return None, None

        data = response.json()
        config = data.get("config", {})

        # Get version information
        version = config.get("version")

        # Check versionIri first (preferred source)
        version_iri = config.get("versionIri")
        if version_iri and _url_exists(version_iri):
            # If we have a versionIri and it exists, use it
            # Extract version from IRI if not already provided
            if not version:
                version = _extract_version_from_iri(version_iri)
            return version_iri, version

        # Fall back to fileLocation if available
        file_location = config.get("fileLocation")
        if file_location and _url_exists(file_location):
            if not version and version_iri:
                # even when version_iri is not accessible, we can still extract the version, for example: pw
                version = _extract_version_from_iri(version_iri)
            return file_location, version

        # No valid URLs found
        return None, None

    except requests.RequestException:
        return None, None



import base64
import hashlib
import json
import os
import secrets
import string
from pathlib import Path

import pandas as pd
from github import Github
from rich.progress import track

BASE_BFX_PIPELINES_PATH = "./scripts/bfxpipelines_info"


def base62(n_char: int) -> str:
    """Like nanoid without hyphen and underscore."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def to_b64_str(bstr: bytes) -> str:
    b64 = base64.urlsafe_b64encode(bstr).decode().strip("=")
    return b64


def hash_str(s: str) -> str:
    bstr = s.encode("utf-8")
    # as we're truncating at a short length, we choose md5 over sha512
    return to_b64_str(hashlib.md5(bstr).digest())


def hash_id(input_id: str | None = None, *, n_char: int) -> str:
    if input_id is None:
        return base62(n_char=n_char)
    else:
        return hash_str(input_id)[:n_char].replace("_", "0").replace("-", "0")


def generate_nf_core_pipelines_info() -> None:
    """Generates a json file that contains all required pipelines information by querying the nf-core Github org."""
    gh_login = Github(os.getenv("GITHUB_TOKEN"))
    nf_core_org = gh_login.get_organization("nf-core")
    blacklist = ["cookiecutter", "tools"]
    nf_core_pipelines = {}

    for repo in track(
        nf_core_org.get_repos(),
        description="Fetching information from nf-core repositories...",
    ):
        if "pipeline" in list(repo.get_topics()):
            if repo.name in blacklist:
                continue

            for version in repo.get_releases():
                actual_version = (
                    version.tag_name if len(version.tag_name) >= 1 else "pre-release"
                )
                pipeline_name = f"{repo.name} v{actual_version}"
                underscore_pipeline_name = (
                    pipeline_name.replace(" ", "_").replace(".", "_").replace("-", "_")
                )

                nf_core_pipelines[underscore_pipeline_name] = {
                    "id": hash_id(pipeline_name, n_char=12),
                    "name": f"nf-core {pipeline_name}",
                    "versions": actual_version,
                    "reference": repo.html_url,
                }

    with open(f"{BASE_BFX_PIPELINES_PATH}/nf_core_pipelines.json", "w") as f:
        json_data = json.dumps(nf_core_pipelines, indent=4)
        f.write(json_data)


def merge_json_files(pipelines_folder_path: str | Path, output_path: str) -> None:
    """Merge all JSON files in a folder and write the merged data to a new JSON file.

    Args:
        pipelines_folder_path: Path to the folder containing the JSON files.
        output_path: Path to the output JSON file.
    """
    pipelines_folder_path = Path(pipelines_folder_path)
    file_paths = list(pipelines_folder_path.glob("*.json"))

    pipeline_json: dict = {}

    for file_path in file_paths:
        with open(file_path) as f:
            if not str(file_path).endswith("bfxpipelines.json"):
                pipelines_info = json.load(f)
                pipeline_json = {**pipeline_json, **pipelines_info}

    with open(output_path, "w") as f:
        json.dump(pipeline_json, f, indent=4)


def write_parquet_file(bfxpipelines_json: str, output_path: str) -> None:
    """Takes a bfxpipelines.json file as generated from merge_json_files and writes a corresponding parquet file."""
    with open(bfxpipelines_json) as f:
        data = json.load(f)

    df = pd.DataFrame(data).transpose()
    df.drop("versions", inplace=True, axis=1)
    df.rename(columns={"id": "ontology_id"}, inplace=True)
    df.set_index("ontology_id", inplace=True, drop=True)
    df.to_parquet(output_path)


generate_nf_core_pipelines_info()
merge_json_files(
    pipelines_folder_path=BASE_BFX_PIPELINES_PATH,
    output_path=f"{BASE_BFX_PIPELINES_PATH}/bfxpipelines.json",
)
write_parquet_file(
    bfxpipelines_json=f"{BASE_BFX_PIPELINES_PATH}/bfxpipelines.json",
    output_path=f"{BASE_BFX_PIPELINES_PATH}/bfxpipelines.parquet",
)



import bionty.base as bt_base
import pandas as pd


def test_mondo_disease_inspect_name():
    df = pd.DataFrame(
        index=[
            "supraglottis cancer",
            "alexia",
            "trigonitis",
            "paranasal sinus disorder",
            "This disease does not exist",
        ]
    )

    ds = bt_base.Disease(source="mondo", version="2025-06-03")
    inspected_df = ds.inspect(df.index, field=ds.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_doid_disease_inspect_ontology_id():
    df = pd.DataFrame(
        index=[
            "DOID:0001816",
            "DOID:0002116",
            "DOID:5547",
            "DOID:5551",
            "This disease does not exist",
        ]
    )

    ds = bt_base.Disease(source="doid")
    inspected_df = ds.inspect(df.index, field=ds.ontology_id, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_icd_9_disease_inspect_name():
    df = pd.DataFrame(
        index=[
            "Cholera d/t vib cholerae",
            "Typhoid fever",
            "Mult gest-plac/sac NOS",
            "Paratyphoid fever a",
            "This disease does not exist",
        ]
    )

    ds = bt_base.Disease(source="icd", version="icd-9-2011")
    inspected_df = ds.inspect(df.index, field=ds.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_icd_10_disease_inspect_name():
    df = pd.DataFrame(
        index=[
            "Vaping-related disorder",
            "COVID-19",
            "Typhoid fever with heart involvement",
            "Typhoid fever, unspecified",
            "This disease does not exist",
        ]
    )

    ds = bt_base.Disease(source="icd", version="icd-10-2020")
    inspected_df = ds.inspect(df.index, field=ds.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_icd_11_disease_inspect_name():
    df = pd.DataFrame(
        index=[
            "Certain infectious or parasitic diseases",
            "Cholera",
            "Intestinal infection due to other Vibrio",
            "Gastroenteritis or colitis of infectious origin",
            "This disease does not exist",
        ]
    )

    ds = bt_base.Disease(source="icd", version="icd-11-2023")
    inspected_df = ds.inspect(df.index, field=ds.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



# Generated by Django 5.2 on 2025-06-03 09:46

import django.db.models.deletion
import lamindb.base.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0053_recordcellline_cellline_records_recordcellmarker_and_more"),
        ("lamindb", "0104_alter_branch_uid"),
    ]

    operations = [
        migrations.AlterField(
            model_name="cellline",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="cellline",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="cellmarker",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="cellmarker",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="celltype",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="celltype",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="developmentalstage",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="developmentalstage",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="disease",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="disease",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="ethnicity",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="ethnicity",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="experimentalfactor",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="experimentalfactor",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="organism",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="organism",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="pathway",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="pathway",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="phenotype",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="phenotype",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="protein",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="protein",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="tissue",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="tissue",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, NamedTuple

import pandas as pd
from lamin_utils import logger

from bionty.base._public_ontology import PublicOntology
from bionty.base._settings import settings
from bionty.base.dev._doc_util import _doc_params
from bionty.base.dev._handle_sources import LAMINDB_INSTANCE_LOADED
from bionty.base.dev._io import s3_bionty_assets

from ._organism import Organism
from ._shared_docstrings import doc_entites

if TYPE_CHECKING:
    from collections.abc import Iterable


class MappingResult(NamedTuple):
    """Result of mapping legacy Ensembl gene IDs to current IDs.

    Attributes:
        mapped: Dictionary of successfully mapped old ensembl IDs to new ensembl IDs
        ambiguous: Dictionary of ambiguously mapped old ensembl IDs to new ensembl IDs - when
            a legacy ID maps to multiple current IDs
        unmapped: List of old ensembl IDs that couldn't be found in the current version
    """

    mapped: dict[str, str]
    ambiguous: dict[str, list[str]]
    unmapped: list[str]


@_doc_params(doc_entities=doc_entites)
class Gene(PublicOntology):
    """Gene.

    1. Ensembl
    https://www.ensembl.org/

    Args:
        {doc_entities}

    Notes:
        Biotypes: https://www.ensembl.org/info/genome/genebuild/biotypes.html
        Gene Naming: https://www.ensembl.org/info/genome/genebuild/gene_names.html
    """

    def __init__(
        self,
        organism: Literal["human", "mouse", "saccharomyces cerevisiae"] | None = None,
        source: Literal["ensembl"] | None = None,
        version: Literal[
            "release-112",
            "release-111",
            "release-110",
            "release-109",
        ]
        | None = None,
        **kwargs,
    ):
        self.taxa = kwargs.pop("taxa", "vertebrates")
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            ols_supported=False,
            **kwargs,
        )

    def _load_df(self):
        if self.source == "ensembl":
            df = super()._load_df()
            if df.empty:
                # Load the Ensembl gene table
                df = EnsemblGene(
                    organism=self._organism, version=self._version, taxa=self.taxa
                ).download_df()
                df.to_parquet(self._local_parquet_path)
            return df
        return super()._load_df()

    # TODO: generalize this to all sources
    def register_source_in_lamindb(self):
        """Register the source in lamindb."""
        if not self._df.empty and LAMINDB_INSTANCE_LOADED:
            import bionty as bt

            # Register the source in lamindb
            source_kwargs = {
                "entity": "bionty.Gene",
                "organism": self._organism,
                "name": self.source,
                "version": self._version,
            }
            source = bt.Source.filter(**source_kwargs).one_or_none()
            if source is None:
                source = bt.Source(**source_kwargs, description="Ensembl").save()
                bt.Gene.add_source(source, df=self._df)
            else:
                logger.warning("source already exists!")

    def map_legacy_ids(self, values: Iterable) -> MappingResult:
        """Convert legacy ids to current IDs.

        This only works if the legacy ensembl ID has a reference to the current ensembl ID.
        It is possible that the HGNC IDs are identical, and the genomic locations are very close,
        but a mapping may still be absent.

        Args:
            values: Legacy ensemble gene IDs of any version

        Example::

            import bionty as bt

            gene = bt.base.Gene()
            gene.map_legacy_genes(["ENSG00000260150", "ENSG00000260587"])
        """
        if self.source != "ensembl":
            raise NotImplementedError
        if isinstance(values, str):
            values = [values]
        ensembl = EnsemblGene(organism=self.organism, version=self.version)
        legacy_df_filename = f"df-legacy_{self.organism}__{self.source}__{self.version}__{self.__class__.__name__}.parquet"
        legacy_df_localpath = settings.dynamicdir / legacy_df_filename
        s3_bionty_assets(
            filename=legacy_df_filename,
            assets_base_url="s3://bionty-assets",
            localpath=legacy_df_localpath,
        )
        try:
            results = pd.read_parquet(legacy_df_localpath)
        except FileNotFoundError:
            raise NotImplementedError from None
        results = results[results.old_stable_id.isin(values)].copy()
        return ensembl._process_convert_result(results, values=values)


class EnsemblGene:
    def __init__(
        self,
        organism: str,
        version: str,
        taxa: Literal[
            "vertebrates", "bacteria", "fungi", "metazoa", "plants", "all"
        ] = "vertebrates",
    ) -> None:
        """Ensembl Gene using direct PyMySQL connection.

        Args:
            organism: Name of the organism
            version: Name of the ensembl DB version, e.g. "release-110"
            taxa: The taxa of the organism to fetch genes for.
        """
        self._import()

        self._organism = (
            Organism(source="ensembl", version=version, taxa=taxa)  # type:ignore
            .lookup()
            .dict()
            .get(organism)
        )

        # Determine port based on taxa
        self._port = 4157 if taxa == "plants" else 3306
        self._host = "ensembldb.ensembl.org"
        self._db = self._organism.core_db
        self._conn = None

    def _import(self):
        try:
            # Check for required dependencies
            import pymysql  # type: ignore
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To query from the Ensembl database, please run:\npip install pymysql"
            ) from None

    def _get_connection(self):
        """Get a PyMySQL connection to the Ensembl database."""
        import pymysql

        if self._conn is None or not self._conn.open:
            self._conn = pymysql.connect(
                host=self._host,
                port=self._port,
                user="anonymous",
                password="",
                database=self._db,
            )
        return self._conn

    def _execute_query(self, query: str) -> pd.DataFrame:
        """Execute a SQL query using PyMySQL and return results as DataFrame."""
        conn = self._get_connection()
        try:
            with conn.cursor() as cursor:
                cursor.execute(query)
                columns = [col[0] for col in cursor.description]
                return pd.DataFrame(cursor.fetchall(), columns=columns)
        except Exception as e:
            logger.error(f"Error executing query: {e}")
            logger.error(f"Query: {query}")
            conn.close()
            self._conn = None
            raise e

    def external_dbs(self) -> pd.DataFrame:
        """Get all external database information."""
        return self._execute_query("SELECT * FROM external_db")

    def download_df(
        self, external_db_names: dict[str, str] | None = None
    ) -> pd.DataFrame:
        """Fetch gene table from Ensembl database.

        Args:
            external_db_names: {external database name : df column name}, see `.external_dbs()`
                Default is {"EntrezGene": "ncbi_gene_id"}.
        """
        query_core = """
        SELECT gene.stable_id, xref.display_label, gene.biotype, gene.description, external_synonym.synonym
        FROM gene
        LEFT JOIN xref ON gene.display_xref_id = xref.xref_id
        LEFT JOIN external_synonym ON gene.display_xref_id = external_synonym.xref_id
        """
        # Make sure to always include EntrezGene
        external_db_names = {
            **{"EntrezGene": "ncbi_gene_id"},
            **(external_db_names or {}),
        }
        external_db_names_str = ", ".join(
            [f"'{name}'" for name in external_db_names.keys()]
        )

        query_external = f"""
        SELECT gene.stable_id, object_xref.xref_id, xref.dbprimary_acc, external_db.db_name
        FROM gene
        LEFT JOIN object_xref ON gene.gene_id = object_xref.ensembl_id
        LEFT JOIN xref ON object_xref.xref_id = xref.xref_id
        LEFT JOIN external_db ON xref.external_db_id = external_db.external_db_id
        WHERE object_xref.ensembl_object_type = 'Gene' AND external_db.db_name IN ({external_db_names_str})
        """

        # Query for the basic gene annotations
        results_core = self._execute_query(query_core)
        logger.info("fetching records from the core DB...")

        # Aggregate metadata based on ensembl stable_id
        results_core_group = results_core.groupby("stable_id").agg(
            {
                "display_label": "first",
                "biotype": "first",
                "description": "first",
                "synonym": lambda x: "|".join([i for i in set(x) if i is not None]),
            }
        )

        # Query for external ids
        results_external = self._execute_query(query_external)
        logger.info("fetching records from the external DBs...")

        def add_external_db_column(
            df: pd.DataFrame, ext_db: str, df_col: str
        ) -> pd.DataFrame:
            # Filter by the external database name
            ext = (
                results_external[results_external["db_name"] == ext_db]
                .drop_duplicates(["stable_id", "dbprimary_acc"])
                .drop(columns=["xref_id", "db_name"])
            )
            ext.rename(columns={"dbprimary_acc": df_col}, inplace=True)
            ext = ext.set_index("stable_id")

            # Check for duplicates
            dup = ext[ext.index.duplicated(keep=False)]
            if dup.shape[0] > 0:
                logger.warning(
                    f"duplicated #rows ensembl_gene_id with {df_col}: {dup.shape[0]}"
                )

            # Merge with existing dataframe
            df_merge = df.merge(ext, left_index=True, right_index=True, how="outer")
            return df_merge

        # Add all external database columns (including EntrezGene)
        df = results_core_group.copy()
        for ext_db, df_col in external_db_names.items():
            df = add_external_db_column(df=df, ext_db=ext_db, df_col=df_col)

        # Finalize the dataframe
        df = df.reset_index().rename(
            columns={
                "stable_id": "ensembl_gene_id",
                "display_label": "symbol",
                "synonym": "synonyms",
            }
        )

        # Define the columns we want to prioritize (they'll appear first in the dataframe)
        priority_columns = [
            "ensembl_gene_id",
            "symbol",
            "ncbi_gene_id",
            "biotype",
            "description",
            "synonyms",
        ]
        # Create a new ordered list with priority columns first, then any remaining columns
        all_columns = priority_columns + [
            col for col in df.columns if col not in priority_columns
        ]
        df_res = df[all_columns].copy()

        # Remove rows with null ensembl_gene_id
        df_res = df_res[df_res["ensembl_gene_id"].notna()]

        # Separate IDs: ensembl_gene_id for ENS-prefixed, stable_id for others
        logger.debug("Separating IDs based on ENS prefix")

        # Create a mask for ENS-prefixed IDs
        is_ens = df_res["ensembl_gene_id"].str.startswith("ENS")
        # If there are any non-ENS IDs, add a stable_id column
        if not all(is_ens):
            logger.warning(
                f"Found {(~is_ens).sum()} IDs without ENS prefix, moving these to stable_id column."
            )
            # Add stable_id column if it doesn't exist
            if "stable_id" not in df_res.columns:
                df_res.insert(0, "stable_id", None)
            # Move non-ENS IDs to stable_id column
            df_res.loc[~is_ens, "stable_id"] = df_res.loc[~is_ens, "ensembl_gene_id"]
            # Clear the ensembl_gene_id for these rows
            df_res.loc[~is_ens, "ensembl_gene_id"] = None

        # Sort by stable_id first (if it exists and has values), then by ensembl_gene_id
        if "stable_id" in df_res.columns and df_res["stable_id"].notna().any():
            df_res = df_res.sort_values(
                ["stable_id", "ensembl_gene_id"], na_position="last"
            ).reset_index(drop=True)
        else:
            df_res = df_res.sort_values("ensembl_gene_id").reset_index(drop=True)

        if "description" in df_res.columns:
            df_res["description"] = df_res["description"].str.replace(
                r"\[.*?\]", "", regex=True
            )
        logger.important(f"downloaded Gene table containing {df_res.shape[0]} entries.")
        return df_res

    def download_legacy_ids_df(
        self, df: pd.DataFrame, col: str | None = None
    ) -> pd.DataFrame:
        """Download legacy Ensembl gene IDs for the current IDs.

        Args:
            df: DataFrame containing Ensembl gene IDs
            col: Column name in df that contains the Ensembl gene IDs

        Returns:
            DataFrame containing mapping between current and legacy IDs
        """
        col = "ensembl_gene_id" if col is None else col

        # Filter out None/NaN values to prevent SQL errors
        valid_ids = df[df[col].notna()][col].tolist()

        # Handle empty list case
        if not valid_ids:
            logger.warning(
                "No valid IDs found in the specified column. Returning empty DataFrame."
            )
            return pd.DataFrame()

        # Format for SQL IN clause - need to add quotes around each ID
        formatted_ids = "(" + ", ".join(f"'{id}'" for id in valid_ids) + ")"

        # Construct and execute query
        query = f"""
            SELECT * FROM stable_id_event
            JOIN mapping_session USING (mapping_session_id)
            WHERE type = 'gene'
            AND new_stable_id IN {formatted_ids}
            AND score > 0
            AND old_stable_id != new_stable_id
        """
        # Execute the query and fetch results
        try:
            results = self._execute_query(query)
            logger.info(f"Downloaded {len(results)} legacy ID mappings")
            return results
        except Exception as e:
            logger.error(f"Error querying legacy IDs: {e}")
            # Return an empty DataFrame rather than failing
            return pd.DataFrame()

    def map_legacy_ids(self, values: Iterable, df: pd.DataFrame) -> MappingResult:
        """Maps legacy gene IDs to current Ensembl gene IDs.

        Takes legacy gene IDs and maps them to current Ensembl IDs by querying the Ensembl database.
        Returns mapping results categorized as unique mappings,
        ambiguous mappings (one legacy ID maps to multiple current IDs), and unmapped IDs.

        Args:
            values: Single gene ID string or iterable of gene ID strings to map
            df: DataFrame containing current Ensembl gene IDs in 'ensembl_gene_id' column

        Example::

            map_legacy_ids(['ENSG00000139618'], df)
            #> MappingResult(
            #>     mapped={'ENSG00000139618': 'ENSG00000012048'},
            #>     ambiguous={},
            #>     unmapped=[],
            )
        """
        # Handle single string input
        if isinstance(values, str):
            values = [values]

        # Convert values to list if it's another iterable type
        values_list = list(values)

        # Format legacy gene IDs for SQL
        legacy_genes = "(" + ", ".join(f"'{gene}'" for gene in values_list) + ")"

        # Filter out None/NaN values from current IDs
        valid_current_ids = df[df["ensembl_gene_id"].notna()][
            "ensembl_gene_id"
        ].tolist()

        # Handle empty list case
        if not valid_current_ids:
            logger.warning(
                "No valid current IDs found. Returning empty mapping result."
            )
            return MappingResult(mapped={}, ambiguous={}, unmapped=values_list)

        # Format current IDs for SQL
        current_ids_sql = "(" + ", ".join(f"'{id}'" for id in valid_current_ids) + ")"

        try:
            # Query the ensembl database
            query = f"""
                SELECT * FROM stable_id_event
                JOIN mapping_session USING (mapping_session_id)
                WHERE type = 'gene'
                AND old_stable_id IN {legacy_genes}
                AND new_stable_id IN {current_ids_sql}
                AND old_stable_id != new_stable_id
            """

            results = self._execute_query(query)
            return self._process_convert_result(results, values_list)
        except Exception as e:
            logger.error(f"Error mapping legacy IDs: {e}")
            return MappingResult(mapped={}, ambiguous={}, unmapped=values_list)

    def _process_convert_result(
        self,
        results: pd.DataFrame,
        values: Iterable,
    ) -> MappingResult:
        """Process the results of a legacy ID mapping query.

        Args:
            results: DataFrame with mapping results
            values: Original input values to map

        Returns:
            A MappingResult object with mapped, ambiguous, and unmapped entries
        """
        # Unique mappings
        mapper = (
            results.drop_duplicates(["old_stable_id"], keep=False)
            .set_index("old_stable_id")["new_stable_id"]
            .to_dict()
        )

        # Ambiguous mappings (one legacy ID maps to multiple current IDs)
        ambiguous = (
            results[~results["old_stable_id"].isin(mapper)][
                ["old_stable_id", "new_stable_id"]
            ]
            .groupby("old_stable_id", group_keys=False)["new_stable_id"]
            .apply(list)
            .to_dict()
        )

        # Unmappable IDs (input IDs not found in results)
        unmapped = set(values).difference(results["old_stable_id"])

        return MappingResult(
            mapped=mapper, ambiguous=ambiguous, unmapped=list(unmapped)
        )

    def __del__(self):
        """Clean up database connection when the object is destroyed."""
        if hasattr(self, "_conn") and self._conn is not None and self._conn.open:
            self._conn.close()



# Generated by Django 5.2 on 2025-01-27 07:22

import django.db.models.deletion
import lamindb.base.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0048_lamindbv1_part6"),
    ]

    operations = [
        migrations.AlterField(
            model_name="schemacellmarker",
            name="cellmarker",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_schema",
                to="bionty.cellmarker",
            ),
        ),
        migrations.AlterField(
            model_name="schemacellmarker",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_cellmarker",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="schemagene",
            name="gene",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_schema",
                to="bionty.gene",
            ),
        ),
        migrations.AlterField(
            model_name="schemagene",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_gene",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="schemapathway",
            name="pathway",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_schema",
                to="bionty.pathway",
            ),
        ),
        migrations.AlterField(
            model_name="schemapathway",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_pathway",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="schemaprotein",
            name="protein",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_schema",
                to="bionty.protein",
            ),
        ),
        migrations.AlterField(
            model_name="schemaprotein",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_protein",
                to="lamindb.schema",
            ),
        ),
    ]



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, BinaryIO

import pandas as pd
from lamin_utils import logger

if TYPE_CHECKING:
    from pathlib import Path


def import_pronto():
    """Import pronto if not already imported."""
    try:
        import warnings

        import pronto  # type: ignore

        if logger._verbosity <= 3:
            # Only apply if not building docs - leads to 'category is not a class' warnings otherwise
            if "sphinx" not in sys.modules:
                warnings.filterwarnings(
                    "ignore", category=pronto.warnings.ProntoWarning
                )

        return pronto
    except ImportError as exc:
        raise ImportError(
            "pronto is not installed. Please install it with `pip install pronto`."
        ) from exc


# Try to create the real class, fall back to a stub
try:
    pronto = import_pronto()

    class Ontology(pronto.Ontology):  # type:ignore
        """Interface with ontologies via pronto.

        Also see: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html

        Args:
            handle: Path to an ontology source file.
            import_depth: The maximum depth of imports to resolve in the ontology tree.
            timeout: The timeout in seconds to use when performing network I/O.
            threads: The number of threads to use when parsing.
            url: The url of the ontology.
            prefix: Dev only -> prefix for get_term.
        """

        def __init__(
            self,
            handle: str | Path | BinaryIO | None = None,
            import_depth: int = -1,
            timeout: int = 100,
            threads: int | None = None,
            prefix: str = "",
        ) -> None:
            self._prefix = prefix
            try:
                logger.debug(f"Attempting to parse ontology file: {handle}")
                super().__init__(
                    handle=handle,
                    import_depth=import_depth,
                    timeout=timeout,
                    threads=threads,
                )
                logger.debug("Ontology parsing successful")
            except Exception as e:
                if "ParseError" in str(type(e).__name__) or "unclosed token" in str(e):
                    logger.warning(f"Skipping corrupted ontology file: {handle}")
                    raise ValueError(f"Corrupted ontology file: {str(e)}") from e
                else:
                    raise

        def get_term(self, term):
            """Search an ontology by its id."""
            try:
                return super().get_term(term)
            except KeyError:
                return super().get_term(f"{self._prefix}{term.replace(':', '_')}")

        def to_df(
            self,
            source: str | None = None,
            include_rel: str | None = None,
            include_id_prefixes: dict[str, list[str]] | None = None,
        ):
            """Convert `pronto.Ontology` to a DataFrame with columns id, name, parents, definition, and synonyms.

            Args:
                source: The source of the ontology terms to include. If None, all terms are included.
                include_rel: The relationship ID to include when gathering parent relationships.
                    If provided, it extends the superclasses with the specified relationship.
                include_id_prefixes: A dictionary mapping sources to lists of ID prefixes.
                    Only terms with IDs starting with these prefixes will be included for the specified source.

            Returns:
                A DataFrame containing the filtered ontology terms with columns:
                - ontology_id (index): The unique identifier for each term.
                - name: The name of the term.
                - definition: The definition of the term (None if not available).
                - synonyms: A pipe-separated string of exact synonyms (None if no synonyms).
                - parents: A list of parent term IDs, including superclasses and optionally the specified relationship.
            """
            logger.info(f"starting ontology `to_df()` conversion for source: {source}")

            def filter_include_id_prefixes(terms: pronto.ontology._OntologyTerms):  # type:ignore
                if include_id_prefixes and source in list(include_id_prefixes.keys()):
                    logger.info(
                        f"filtering terms by ID prefixes for source {source}: {include_id_prefixes[source]}"
                    )
                    return list(
                        filter(
                            lambda val: any(
                                val.id.startswith(f"{prefix}")
                                for prefix in include_id_prefixes[source]  # type: ignore
                            ),
                            terms,
                        )
                    )
                else:
                    return terms

            if source is not None:
                prefix_list = (
                    include_id_prefixes.get(source)
                    if include_id_prefixes is not None
                    else None
                )
            else:
                prefix_list = None

            logger.info("filtering terms by ID prefixes...")
            filtered_terms = filter_include_id_prefixes(self.terms())

            logger.info("processing individual terms...")
            df_values = []

            processed_count = 0
            log_interval = max(1, len(filtered_terms) // 20)  # log every 5% of terms

            for i, term in enumerate(filtered_terms):
                if i % log_interval == 0 and i > 0:
                    print("  ", end="")
                    logger.info(
                        f"Processed {i}/{len(filtered_terms)} terms ({i / len(filtered_terms) * 100:.1f}%)"
                    )

                # skip terms without id or name
                if (not term.id) or (not term.name):
                    continue

                # term definition text
                definition = (
                    None if term.definition is None else term.definition.title()
                )

                # concatenate synonyms into a string
                synonyms = "|".join(
                    [
                        synonym.description
                        for synonym in term.synonyms
                        if synonym.scope == "EXACT"
                    ]
                )
                if len(synonyms) == 0:
                    synonyms = None  # type:ignore

                # get 1st degree parents and additional relatonships (include_rel such as 'part_of')
                superclasses = [
                    superclass.id
                    for superclass in term.superclasses(
                        distance=1, with_self=False
                    ).to_set()
                ]

                if include_rel is not None:
                    if include_rel in [i.id for i in term.relationships]:
                        superclasses.extend(
                            [
                                superclass.id
                                for superclass in term.objects(
                                    self.get_relationship(include_rel)
                                )
                            ]
                        )

                if prefix_list is not None:
                    superclasses = [
                        superclass
                        for superclass in superclasses
                        if superclass.startswith(tuple(prefix_list))
                    ]

                df_values.append(
                    (term.id, term.name, definition, synonyms, superclasses)
                )
                processed_count += 1

            logger.success(f"processed {processed_count} terms")

            df = pd.DataFrame(
                df_values,
                columns=["ontology_id", "name", "definition", "synonyms", "parents"],
            )
            logger.success(f"created DataFrame with {len(df)} rows")

            df["ontology_id"] = [
                i.replace(self._prefix, "").replace("_", ":") for i in df["ontology_id"]
            ]
            df["parents"] = [
                [j.replace(self._prefix, "").replace("_", ":") for j in i]
                for i in df["parents"]
            ]

            # needed to avoid erroring in .lookup()
            df["name"] = df["name"].fillna("")
            logger.success("conversion completed")

            return df.set_index("ontology_id")

except ImportError:
    # Create a stub class that raises an error when instantiated
    class Ontology:  # type:ignore
        def __init__(self, *args, **kwargs):
            raise ImportError(
                "pronto is not installed. Please install it with `pip install pronto`."
            )



# Generated by Django 5.2 on 2025-07-26 15:56

import lamindb.base.fields
from django.db import migrations

import bionty.uids


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0056_alter_recordtissue_record"),
    ]

    operations = [
        migrations.AlterField(
            model_name="cellline",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="cellline",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="cellline",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="cellmarker",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="cellmarker",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="cellmarker",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.cellmarker,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="celltype",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="celltype",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="celltype",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="developmentalstage",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="developmentalstage",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="developmentalstage",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="disease",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="disease",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="disease",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="ethnicity",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ethnicity",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ethnicity",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="experimentalfactor",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="experimentalfactor",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="experimentalfactor",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="ncbi_gene_ids",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.gene,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="organism",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="organism",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="organism",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="pathway",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="pathway",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="pathway",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="phenotype",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="phenotype",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="phenotype",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="protein",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="protein",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="protein",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.protein,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="tissue",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="tissue",
            name="synonyms",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="tissue",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=bionty.uids.ontology,
                max_length=8,
                unique=True,
            ),
        ),
    ]
